/* Copyright 2020-2025, Michele Santullo
 * This file is part of memoserv.
 *
 * Memoserv is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Memoserv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Memoserv.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "savegame_db.hpp"
#include "memcard/alignment.hpp"
#include <algorithm>
#include <ciso646>
#include <iterator>
#include <utility>
#include <tuple>

namespace duck {
namespace {
} //unnamed namespace

BlockEntry::BlockEntry (
	std::unique_ptr<uint8_t[]>&& data,
	std::size_t count
) :
	data(std::move(data)),
	block_count(count)
{
}

std::string_view BlockEntry::identifier() const {
	return make_block<true>(0).identifier();
}

std::string BlockEntry::title() const {
	return make_block<true>(0).title();
}

SavegameDb::SavegameDb() = default;
SavegameDb::~SavegameDb() noexcept = default;

template <bool Const>
unsigned int SavegameDb::add (const std::vector<mc::psx::BasicBlock<Const> >& blocks) {
	constexpr auto block_size = mc::psx::BasicBlock<Const>::BlockByteSize;
	constexpr auto frame_size = mc::psx::BasicFrame<Const>::Size;
	assert(not blocks.empty());

	const auto block_count = blocks.size();
	assert(block_count == blocks.front().block_count());

	static_assert(
		frame_size % mc::psx::Block::DataAlignment == 0,
		"Unexpected alignment, alignment in this class will be wrong if you ignore this error"
	);

	auto mem = mc::psx::make_unique_for_block_aligned(frame_size + block_size * block_count);

	std::uint8_t* const ptr = mc::psx::to_block_aligned(mem.get());
	uint8_t* dest = std::copy(blocks.front().toc().cbegin(), blocks.front().toc().cend(), ptr);
	assert(mc::psx::is_block_aligned(dest));
	assert(std::distance(ptr, dest) == frame_size);
	mc::psx::Frame toc(ptr, 0);

	for (const auto& block : blocks) {
		assert(block_count == block.block_count() or block.block_count() == 0);
		uint8_t* const curr_block_ptr = dest;
		dest = std::copy(block.data(), block.data() + block_size, dest);
	}

	const std::size_t new_id = m_next_id++;
	m_saves.emplace(
		std::piecewise_construct,
		std::forward_as_tuple(new_id),
		std::forward_as_tuple(std::move(mem), block_count)
	);
	m_block_count += block_count;
	return static_cast<unsigned int>(m_saves.size() - 1u);
}

std::size_t SavegameDb::savegame_count() const {
	return m_saves.size();
}

std::size_t SavegameDb::block_count() const {
	return m_block_count;
}

auto SavegameDb::begin() -> iterator {
	return {m_saves.begin()};
}

auto SavegameDb::end() -> iterator {
	return {m_saves.end()};
}

template unsigned int SavegameDb::add<true> (const std::vector<mc::psx::BasicBlock<true> >&);
template unsigned int SavegameDb::add<false> (const std::vector<mc::psx::BasicBlock<false> >&);
} //namespace duck
