/* Copyright 2020-2025, Michele Santullo
 * This file is part of memoserv.
 *
 * Memoserv is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Memoserv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Memoserv.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "memcard/alignment.hpp"
#include "memcard/block.hpp"
#include <cstdint>
#include <vector>
#include <memory>
#include <string_view>
#include <string>
#include <map>
#include <utility>
#include <iterator>
#include <cassert>
#include <ciso646>
#include <type_traits>

namespace duck {
struct BlockEntry {
	BlockEntry (
		std::unique_ptr<uint8_t[]>&& data,
		std::size_t count
	);

	template <bool C>
	mc::psx::BasicBlock<C> make_block (std::size_t index) const;
	std::string_view identifier() const;
	std::string title() const;
	std::uint8_t* data_ptr() const;

	std::unique_ptr<std::uint8_t[]> data;
	std::size_t block_count;
};

class SavegameDb {
	typedef std::map<unsigned int, BlockEntry> SavegameMap;
public:
	template <typename It>
	class BlockIterator {
		It m_it;
		std::size_t m_pos_in_it {0};

	public:
		typedef typename std::iterator_traits<It>::difference_type difference_type;
		typedef mc::psx::BasicBlock<std::is_const<SavegameMap::mapped_type>::value> value_type;
		typedef std::add_pointer<value_type>::type pointer;
		typedef value_type reference;
		typedef mc::psx::ConstBlock const_reference;
		typedef typename std::iterator_traits<It>::iterator_category iterator_category;

		BlockIterator() = default;
		BlockIterator(const BlockIterator&) = default;
		BlockIterator(BlockIterator&&) = default;
		BlockIterator (It&& it) : m_it(std::move(it)) {}
		BlockIterator (const It& it) : m_it(it) {}

		bool operator!=(const BlockIterator& other) const
			{return m_it!=other.m_it or m_pos_in_it!=other.m_pos_in_it;}
		bool operator==(const BlockIterator& other) const
			{return m_it==other.m_it and m_pos_in_it==other.m_pos_in_it;}

		BlockIterator& operator++();
		BlockIterator operator++(int);
		reference operator*();
		const_reference operator*() const;
	};

	typedef BlockIterator<SavegameMap::iterator> iterator;
	typedef BlockIterator<SavegameMap::const_iterator> const_iterator;

	SavegameDb();
	~SavegameDb() noexcept;

	template <bool Const>
	unsigned int add (const std::vector<mc::psx::BasicBlock<Const> >& blocks);
	std::size_t savegame_count() const;
	std::size_t block_count() const;

	std::size_t size() const { return block_count(); }
	iterator begin();
	iterator end();

private:
	SavegameMap m_saves;
	std::size_t m_next_id{0};
	std::size_t m_block_count{0};
};

template <bool C>
mc::psx::BasicBlock<C> BlockEntry::make_block (std::size_t index) const {
	using mc::psx::to_block_aligned;
	using mc::psx::Frame;
	using FrameType = typename mc::psx::BasicBlock<C>::FrameType;

	assert(this->block_count > 0);
	assert(this->block_count > index);

	constexpr std::size_t frame_index = 0;
	return {
		to_block_aligned(this->data.get(), index, Frame::Size),
		FrameType{to_block_aligned(this->data.get()), frame_index},
		index
	};
}

template <typename It>
SavegameDb::BlockIterator<It>& SavegameDb::BlockIterator<It>::operator++() {
	const BlockEntry& entry = m_it->second;
	++m_pos_in_it;
	if (m_pos_in_it >= entry.block_count) {
		++m_it;
		m_pos_in_it = 0;
	}
	return *this;
}

template <typename It>
SavegameDb::BlockIterator<It> SavegameDb::BlockIterator<It>::operator++(int) {
	auto retval = *this;
	this->operator++();
	return retval;
}

template <typename It>
auto SavegameDb::BlockIterator<It>::operator*() -> reference {
	BlockEntry& entry = m_it->second;
	return entry.make_block<value_type::IsConst>(m_pos_in_it);
}

template <typename It>
auto SavegameDb::BlockIterator<It>::operator*() const -> const_reference {
	BlockEntry& entry = m_it->second;
	return entry.make_block<value_type::IsConst>(m_pos_in_it);
}
} //namespace duck
