/* Copyright 2016-2025 Michele Santullo
 * This file is part of "duckhandy".
 *
 * "duckhandy" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * "duckhandy" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "duckhandy".  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef id83CE9FAB90684250B98EA584C548E13F
#define id83CE9FAB90684250B98EA584C548E13F

#include "bitfield_pack.hpp"
#include <cstdint>
#include <utility>
#include <climits>
#include <type_traits>
#include <cassert>
#include <concepts>

namespace dhandy {
namespace detail {

//This monster expression is just the one-liner version of this
//implementation of log2:
//unsigned int v;	         // 32-bit value to find the log2 of
//register unsigned int r; // result of log2(v) will go here
//register unsigned int shift;
//
//r =     (v > 0xFFFF) << 4; v >>= r;
//shift = (v > 0xFF  ) << 3; v >>= shift; r |= shift;
//shift = (v > 0xF   ) << 2; v >>= shift; r |= shift;
//shift = (v > 0x3   ) << 1; v >>= shift; r |= shift;
//                                        r |= (v >> 1);
//see https://graphics.stanford.edu/~seander/bithacks.html#IntegerLog for
//an explanation of that.
//This is a shorter version of it but it only works if the input is a power
//of 2:
//(((((NUM&0xAAAAAAAA)!=0)|((NUM&0xFFFF0000)!=0)<<4)|((NUM&0xFF00FF00)!=0)<<
//3)|((NUM&0xF0F0F0F0)!=0)<<2)|((NUM&0xCCCCCCCC)!=0)<<1
//Source for that is also available at the page linked above.
template <std::uint32_t NUM> constexpr std::size_t pointer_unused_bit_count =
	((((((NUM > 0xFFFF) << 4) | (((NUM >> ((NUM > 0xFFFF) << 4)) > 0xFF) <<
	3)) | ((((NUM >> ((NUM > 0xFFFF) << 4)) >> (((NUM >> ((NUM > 0xFFFF) <<
	4)) > 0xFF) << 3)) > 0xF) << 2)) | (((((NUM >> ((NUM > 0xFFFF) << 4)) >>
	(((NUM >> ((NUM > 0xFFFF) << 4)) > 0xFF) << 3)) >> ((((NUM >> ((NUM >
	0xFFFF) << 4)) >> (((NUM >> ((NUM > 0xFFFF) << 4)) > 0xFF) << 3)) > 0xF)
	<< 2)) > 0x3) << 1)) | (((((NUM >> ((NUM > 0xFFFF) << 4)) >> (((NUM >> (
	(NUM > 0xFFFF) << 4)) > 0xFF) << 3)) >> ((((NUM >> ((NUM > 0xFFFF) << 4)
	) >> (((NUM >> ((NUM > 0xFFFF) << 4)) > 0xFF) << 3)) > 0xF) << 2)) >> ((
	(((NUM >> ((NUM > 0xFFFF) << 4)) >> (((NUM >> ((NUM > 0xFFFF) << 4)) >
	0xFF) << 3)) >> ((((NUM >> ((NUM > 0xFFFF) << 4)) >> (((NUM >> ((NUM >
	0xFFFF) << 4)) > 0xFF) << 3)) > 0xF) << 2)) > 0x3) << 1)) >> 1))
;

template <std::size_t... S>
BitfieldPack<
	std::uintptr_t,
	sizeof(std::uintptr_t) * CHAR_BIT - sizeof...(S), (S+1)/(S+1)...
>
guess_bitfield_type (std::index_sequence<S...>);

template <std::size_t PtrAlign>
using BaseBitfieldPackForPackedPointer = decltype(
	guess_bitfield_type(
		std::make_index_sequence<pointer_unused_bit_count<PtrAlign>>()
	)
);

template <std::size_t PtrAlign>
using BaseIntPackForPackedPointer = BitfieldPack<
	std::uintptr_t,
	sizeof(std::uintptr_t) * CHAR_BIT - pointer_unused_bit_count<PtrAlign>,
	pointer_unused_bit_count<PtrAlign>
>;

template <std::size_t PtrAlign, bool BitFlags>
using BaseBitfieldType = std::conditional_t<BitFlags,
	BaseBitfieldPackForPackedPointer<PtrAlign>, BaseIntPackForPackedPointer<PtrAlign>
>;

template <typename T, typename I>
struct PackedPointerSetterGetter {
	template <std::size_t Idx, std::size_t S1, std::size_t S2>
	static void set (dhandy::BitfieldPack<std::uintptr_t, S1, S2>& out, I val) {
		static_assert(0 == Idx);
		dhandy::set<1>(out, val);
	}
	template <std::size_t Idx, std::size_t S1, std::size_t S2>
	static I get (const dhandy::BitfieldPack<std::uintptr_t, S1, S2>& in) {
		static_assert(0 == Idx);
		return dhandy::get<1>(in);
	}
};

template <typename T>
struct PackedPointerSetterGetter<T, bool> {
	template <std::size_t Idx, std::size_t... S>
	static void set (dhandy::BitfieldPack<std::uintptr_t, S...>& out, bool flag) {
		static_assert(Idx < sizeof...(S) - 1, "Index out of range");
		dhandy::set<Idx+1>(out, (flag ? 1u : 0u));
	}
	template <std::size_t Idx, std::size_t... S>
	static bool get (const dhandy::BitfieldPack<std::uintptr_t, S...>& in) {
		static_assert(Idx < sizeof...(S) - 1, "Index out of range");
		return static_cast<bool>(dhandy::get<Idx+1>(in));
	}
};
} //namespace detail

template <typename T, typename I=bool, std::size_t A=alignof(T)>
class PackedPointer : detail::BaseBitfieldType<A, std::is_same<I, bool>::value> {
	static_assert(A >= 2, "Not enough space to pack anything");
	static_assert(sizeof(T*) == sizeof(std::uintptr_t), "Mismatching int/pointer size");
	typedef detail::BaseBitfieldType<A, std::is_same<I, bool>::value> parent_type;
	static_assert(sizeof(parent_type) == sizeof(T*));

	template <std::size_t Idx, typename U, typename J, std::size_t B> friend J get (const PackedPointer<U, J, B>&);
	template <std::size_t Idx, typename U, typename J, std::size_t B> friend void set (PackedPointer<U, J, B>&, J);
public:
	static constexpr std::size_t FlagsCount = parent_type::element_count - 1;
	static constexpr auto max_payload =
		static_cast<std::conditional_t<std::is_same_v<I,bool>, unsigned int, I>>(
			parent_type::template max_at_index<1>
		);
	typedef I int_type;

	PackedPointer();
	PackedPointer (T* ptr);
	PackedPointer (const PackedPointer& other);
	PackedPointer (PackedPointer&& other);
	~PackedPointer() = default;

	operator T*();
	operator const T*() const;
	operator std::uintptr_t() const;
	operator bool() const;

	PackedPointer& operator= (const PackedPointer& other);
	PackedPointer& operator= (PackedPointer&& other);
	PackedPointer& operator= (T* ptr);
	T* operator+ (std::integral auto diff);
	const T* operator+ (std::integral auto diff) const;
	T* operator- (std::integral auto diff);
	const T* operator- (std::integral auto diff) const;
	T* operator->();
	const T* operator->() const;
	T& operator*();
	const T& operator*() const;
	bool operator== (const PackedPointer& other) const;
	bool operator== (T* other) const;
	bool operator!= (const PackedPointer& other) const;
	bool operator!= (T* other) const;

private:
	//static const constexpr uintptr_t PtrMask = static_cast<uintptr_t>(-1) << FlagsCount;
};

template <std::size_t Idx=0, typename T=char, typename I=bool, std::size_t A=0>
inline void set (PackedPointer<T, I, A>& obj, I value) {
	assert(value < (1u << A));
	detail::PackedPointerSetterGetter<T, I>::template set<Idx>(obj, value);
}

template <std::size_t Idx=0, typename T=char, typename I=bool, std::size_t A>
inline I get (const PackedPointer<T, I, A>& obj) {
	return detail::PackedPointerSetterGetter<T, I>::template get<Idx>(obj);
}

template <typename T, typename I, std::size_t A> const constexpr std::size_t PackedPointer<T, I, A>::FlagsCount;
//template <typename T> const constexpr std::uintptr_t PackedPointer<T>::PtrMask;

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>::PackedPointer() :
	PackedPointer(nullptr)
{
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>::PackedPointer (T* ptr) :
	parent_type(0)
{
	(*this) = ptr;
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>::PackedPointer (const PackedPointer& other) :
	parent_type(other)
{
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>::PackedPointer (PackedPointer&& other) :
	parent_type(std::move(other))
{
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>::operator T*() {
	return reinterpret_cast<T*>(get<0>(*static_cast<parent_type*>(this)) << FlagsCount);
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>::operator const T*() const {
	return reinterpret_cast<const T*>(get<0>(*static_cast<const parent_type*>(this)) << FlagsCount);
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>::operator std::uintptr_t() const {
	return get<0>(*static_cast<const parent_type*>(this)) << FlagsCount;
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>::operator bool() const {
	return static_cast<T*>(*this) != nullptr;
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>& PackedPointer<T, I, A>::operator= (const PackedPointer& other) {
	*static_cast<parent_type*>(this) = static_cast<parent_type&>(other);
	return *this;
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>& PackedPointer<T, I, A>::operator= (PackedPointer&& other) {
	*static_cast<parent_type*>(this) = std::move(static_cast<parent_type&>(other));
	return *this;
}

template <typename T, typename I, std::size_t A>
PackedPointer<T, I, A>& PackedPointer<T, I, A>::operator= (T* ptr) {
	assert(reinterpret_cast<std::uintptr_t>(ptr) % A == 0);
	set<0>(*static_cast<parent_type*>(this), reinterpret_cast<std::uintptr_t>(ptr) >> FlagsCount);
	return *this;
}

template <typename T, typename I, std::size_t A>
T* PackedPointer<T, I, A>::operator+ (std::integral auto diff) {
	return static_cast<T*>(*this) + diff;
}

template <typename T, typename I, std::size_t A>
const T* PackedPointer<T, I, A>::operator+ (std::integral auto diff) const {
	return static_cast<const T*>(*this) + diff;
}

template <typename T, typename I, std::size_t A>
T* PackedPointer<T, I, A>::operator- (std::integral auto diff) {
	return static_cast<T*>(*this) + diff;
}

template <typename T, typename I, std::size_t A>
const T* PackedPointer<T, I, A>::operator- (std::integral auto diff) const {
	return static_cast<const T*>(*this) + diff;
}

template <typename T, typename I, std::size_t A>
T* PackedPointer<T, I, A>::operator->() {
	return static_cast<T*>(*this);
}

template <typename T, typename I, std::size_t A>
const T* PackedPointer<T, I, A>::operator->() const {
	return static_cast<T*>(*this);
}

template <typename T, typename I, std::size_t A>
T& PackedPointer<T, I, A>::operator*() {
	return *static_cast<T*>(*this);
}

template <typename T, typename I, std::size_t A>
const T& PackedPointer<T, I, A>::operator*() const {
	return *static_cast<T*>(*this);
}

template <typename T, typename I, std::size_t A>
bool PackedPointer<T, I, A>::operator== (const PackedPointer& other) const {
	return this == &other or static_cast<T*>(*this) == static_cast<T*>(other);
}

template <typename T, typename I, std::size_t A>
bool PackedPointer<T, I, A>::operator== (T* other) const {
	return static_cast<T*>(*this) == other;
}

template <typename T, typename I, std::size_t A>
bool PackedPointer<T, I, A>::operator!= (const PackedPointer& other) const {
	return not this->operator==(other);
}

template <typename T, typename I, std::size_t A>
bool PackedPointer<T, I, A>::operator!= (T* other) const {
	return not this->operator==(other);
}
} //namepace dhandy

#endif
